# see https://github.com/rtfd/CommonMark-py/blob/master/CommonMark/render/html.py
# for the HTML renderer -- this is just a riff on that.

import re

import commonmark
import commonmark.render.renderer


class RawHtmlNotAllowed(ValueError):
    def __init__(self):
        super(RawHtmlNotAllowed, self).__init__("Raw HTML cannot be rendered by the plain text renderer.")


class ListBlock:
    def __init__(self, list_type, start_value, bullet_char):
        self.list_type = list_type
        self.value = start_value
        self.bullet_char = bullet_char
    def __str__(self):
        # Does not cause indentation.
        return ""


class ItemBullet:
    def __init__(self, listblock):
        self.listblock = listblock
        self.emitted = None
    def __str__(self):
        # A bullet is emitted exactly once.
        if not self.emitted:
            if self.listblock.list_type == "bullet":
                self.emitted = self.listblock.bullet_char + " "
            elif self.listblock.list_type == "ordered":
                self.emitted = str(self.listblock.value) + ". "
                self.listblock.value += 1
            else:
                raise ValueError(self.listblock.list_type)
            return self.emitted

        # After that, it is just emitted as indentation.
        else:
            return " " * len(self.emitted)


class PlainTextRenderer(commonmark.render.renderer.Renderer):
    def __init__(self):
        self.setext_heading_chars = ["#", "=", "-"]
        self.block_indent = []

    def emit_intent(self):
        self.lit("".join(str(b) for b in self.block_indent))

    def emit_end_block(self, node):
        # adapted from the HtmlRenderer
        grandparent = node.parent.parent
        if grandparent is not None and grandparent.t == 'list' and  grandparent.list_data['tight']:
            # Within a loose list, don't add a double-newline.
            pass
        else:
            self.emit_intent()
            self.cr()

    def text(self, node, entering=None):
        self.out(node.literal)
    def softbreak(self, node=None, entering=None):
        self.cr()
        self.emit_intent()
    def linebreak(self, node=None, entering=None):
        self.cr()
        self.emit_intent()
    def link(self, node, entering):
        if entering:
            self.link_start = len(self.buf)
        else:
            text = self.buf[self.link_start:]
            if text != node.destination:
                self.lit(" <" + node.destination + ">")
    def image(self, node, entering):
        if entering:
            self.lit('[image]')
        else:
            pass

    def emphstronglevel(self, node):
        if (node.parent.t in ("emph", "strong") and node is node.parent.first_child):
            return self.emphstronglevel(node.parent) + 1
        elif node.prv and node.prv.t in ("emph", "strong"):
            return self.emphstronglevel(node.prv) + 1
        return 0

    def emph(self, node, entering):
        # same symbol entering & exiting, but must alternate between * and _
        # when nested immediately within or following a strong/emph.
        if (self.emphstronglevel(node) % 2) == 0:
            self.lit("*")
        else:
            self.lit("_")
    def strong(self, node, entering):
        # same symbol entering & exiting, but must alternate between * and _
        # when nested immediately within or following a strong/emph.
        if (self.emphstronglevel(node) % 2) == 0:
            self.lit("**")
        else:
            self.lit("__")
    def paragraph(self, node, entering):
        if entering:
            self.emit_intent()
        else:
            self.cr()
            self.emit_end_block(node)
    def heading(self, node, entering):
        if entering:
            self.emit_intent()
            self.heading_start = len(self.buf)
        else:
            if node.level <= len(self.setext_heading_chars):
                heading_len = len(self.buf) - self.heading_start
                if heading_len == 0:
                    # CommonMark requires that the heading still be emitted even if
                    # empty, so fall back to a setext-style heading.
                    self.lit("#" * node.level + " ")
                else:
                    self.cr()
                    self.emit_intent()
                    self.lit(self.setext_heading_chars[node.level-1] * heading_len)
            self.cr()
            self.emit_end_block(node)
    def code(self, node, entering):
        # Just do actual CommonMark here. The backtick string around the literal
        # must have one more backtick than the number of consecutive backticks
        # in the literal.
        backtick_string = "`"
        while backtick_string in node.literal:
            backtick_string += "`"
        self.lit(backtick_string)
        if node.literal.startswith("`") or node.literal == "":
            # Must have space betweet a literal backtick within the code,
            # and if the code is totally empty there must be space between
            # the start and end backticks.
            self.lit(" ")
        self.lit(node.literal) # this is correct as lit() and not out() for CommonMark-compliant output
        if node.literal.endswith("`"):
            self.lit(" ")
        self.lit(backtick_string)
    def code_block(self, node, entering):
        # open code block
        self.emit_intent()
        self.emit_code_block_fence(node.literal, node.info)
        # each line, with indentation; note that the literal is a literal
        # and must not be escaped
        self.emit_intented_literal(node.literal)
        # close code block
        self.emit_intent()
        self.emit_code_block_fence(node.literal)
        self.emit_end_block(node)
    def emit_code_block_fence(self, literal, language=None):
        width = max([len(line.replace("\t", "    ")) for line in literal.split("\n")])
        self.lit("-" * width + "\n")
    def emit_intented_literal(self, literal):
        lines = literal.split("\n")
        while len(lines) > 0 and lines[-1] == "":
            # The parser sometimes includes an extra blank line.
            # Might be a parser bug?
            lines.pop(-1)
            break
        for line in lines:
            self.emit_intent()
            self.lit(line + "\n")
    def thematic_break(self, node, entering):
        self.emit_intent()
        self.lit("-" * 60)
        self.cr()
        self.emit_end_block(node)
    def block_quote(self, node, entering):
        if entering:
            self.block_indent.append("> ")
            self.block_quote_start = len(self.buf)
        else:
            if self.block_quote_start == len(self.buf):
                # If no content, still must emit something.
                self.emit_intent()
                self.cr()
            self.block_indent.pop(-1)
    def list(self, node, entering):
        if entering:
            # We could re-use the bullet character from the input:
            # bullet_char = node.list_data['bullet_char']
            # but for better normalization we'll choose a bullet char by
            # alternating through *, -, and + as we go deeper into levels.
            bullet_level = len(list(filter(lambda b : isinstance(b, ListBlock) and b.list_type == "bullet", self.block_indent)))
            bullet_char = ["*", "-", "+"][bullet_level % 3]
            # TODO #1: Two lists next to each other are distinguished as
            # different if they have either a different bullet (for bulleted
            # lists) or a different delimiter (the "." or ")" after the number,
            # for ordered lists). That distinction might be lost here and
            # would result in two lists being combined into one.
            # TODO #2: A list can be loose or tight, but we don't output them
            # any differently.
            self.block_indent.append(ListBlock(node.list_data['type'], node.list_data['start'], bullet_char))
        else:
            self.block_indent.pop(-1)
            self.emit_end_block(node)
    def item(self, node, entering):
        if entering:
            # Find the ListBlock that was most recently added to self.block_indent.
            parent_list = [b for b in self.block_indent if isinstance(b, ListBlock)][-1]
            self.block_indent.append(ItemBullet(parent_list))
            self.item_start = len(self.buf)
        else:
            if len(self.buf) == self.item_start:
                # Always emit a bullet even if there was no content.
                self.emit_intent()
                self.cr()
            self.block_indent.pop(-1)

    def html_inline(self, node, entering):
        raise RawHtmlNotAllowed()

    def html_block(self, node, entering):
        raise RawHtmlNotAllowed()

    def custom_inline(self, node, entering):
        # copied from the HTML renderer
        if entering and node.on_enter:
            self.lit(node.on_enter)
        elif (not entering) and node.on_exit:
            self.lit(node.on_exit)

    def custom_block(self, node, entering):
        # copied from the HTML renderer
        self.cr()
        if entering and node.on_enter:
            self.lit(node.on_enter)
        elif (not entering) and node.on_exit:
            self.lit(node.on_exit)
        self.emit_end_block(node)

class CommonMarkToCommonMarkRenderer(PlainTextRenderer):
    def __init__(self):
        super(CommonMarkToCommonMarkRenderer, self).__init__()
        self.setext_heading_chars = ["=", "-"]

    def out(self, s):
        # Escape characters that have significance to the CommonMark spec.
        # While all ASCII punctuation can be escaped (http://spec.commonmark.org/0.28/#ascii-punctuation-character),
        # not all ASCII punctuation has significance. Some symbols are
        # significant only in certain context (e.g. start of line) but
        # we don't know the context here.
        always_escape = {
            "`", "~", # fenced code blocks, code spans
            "<", # raw HTML start conditions
            "*", "_", # emphasis and strong emphasis
            "[", "]", # link text
            "<", ">", # link destination, autolink
            "\"", "'", # link title
            # "(", ")", # inline link -- if it were valid, it would have balanced parens so escaping would not be necessary?
            "!", # image
        }

        # Always escape the characters above.
        pattern = "|".join(re.escape(c) for c in always_escape)

        # Escape things that look like character references but aren't.
        pattern += r"|&\w+;|&#[Xx]?[0-9A-Fa-f]+;"

        # Some characters only need escaping at the start of a line, which might
        # include start-of-line characters. But we don't have a way to detect
        # that at the moment. We can just filter out of if it doesn't follow
        # something that can't precede it on a line.
        escape_at_line_start = {
            "---", "___", "***", # thematic break
            "#", # ATX headers
            "=", "-", # setext underline
            "[", # link reference definitions (but not the colon since a stray colon could not be confused here)
            ">", # block quotes
            "-", "+", "*", # bullet list marker
        }
        pattern += "|" + "|".join("(?<![A-Za-z0-9])" + re.escape(c) for c in escape_at_line_start)

        # The ordered list markers need escapes just when they follow a digit.
        pattern += r"|(?<=[0-9])[\.\)]"

        # Escape backslashes if followed by punctuation (other backslashes
        # cannot be for escapes and are treated literally).
        ascii_punctuation_chars = "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~"
        pattern += r"|\\(?=[" + re.escape(ascii_punctuation_chars) +"])"

        # Apply substitutions.
        s = re.sub(pattern, lambda m : "\\" + m.group(0), s, re.M)

        super(CommonMarkToCommonMarkRenderer, self).out(s)

    def linebreak(self, node=None, entering=None):
        self.lit("\\\n")
        self.emit_intent()

    def link(self, node, entering):
        # Determine if the link label and the destination are the same by rendering
        # this node using the plain text renderer. Luckily, if they are the same,
        # the plain text renderer simply emits the label without the destination.
        link_label = PlainTextRenderer().render(node)
        if (link_label == node.destination) or ("mailto:" + link_label.lower().lower() == node.destination.lower()) \
            and re.match(r"[A-Za-z][A-Za-z0-9+\.-]{1,32}:[^<> ]*$", node.destination):
            # Emit an autolink.
            if entering:
                destination = node.destination
                if destination.lower().startswith("mailto:"):
                    destination = destination[7:]
                self.lit("<")
                self.lit(node.destination)
                self.lit(">")
                self.autolink_start = len(self.buf)
            else:
                # kill any content emitted within this node
                self.buf = self.buf[0:self.autolink_start]
        else:
            if entering:
                self.lit("[")
            else:
                self.lit("](")
                # When wrapping the destination with parens, then internal parens must be
                # either well-nested (which is hard to detect) or escaped.
                self.lit(node.destination.replace("(", "\\(").replace(")", "\\)"))
                if node.title:
                    self.lit(" \"")
                    # When wrapping the title in double quotes, internal double quotes
                    # must be escaped.
                    self.lit(node.title.replace("\"", "\\\""))
                    self.lit("\"")
                self.lit(")")

    def image(self, node, entering):
        if entering:
            self.lit("![")
        else:
            self.lit("](")
            # same as link, see above
            self.lit(node.destination.replace("(", "\\(").replace(")", "\\)"))
            if node.title:
                # same as link, see above
                self.lit(" \"")
                self.lit(node.title.replace("\"", "\\\""))
                self.lit("\"")
            self.lit(")")

    def heading(self, node, entering):
        if node.level <= 2:
            # Prefer setext-style heading for levels 1 and 2, because it is
            # the only style that supports multi-line content within it, which
            # we might have (and we don't know at this point).
            super(CommonMarkToCommonMarkRenderer, self).heading(node, entering)
        else:
            # Use ATX-style headings for other levels.
            if entering:
                self.lit("#" * node.level + " ")
                self.block_indent.append(" " * (node.level+1))
            else:
                self.cr()
                self.block_indent.pop(-1)

    def emit_code_block_fence(self, content, info_string=None):
        # Choose a fence string that does not appear in the content
        # of the code block. A fence string can made made up of
        # backticks or tildes, but we'll stick to backticks.
        fence_string = "```"
        while fence_string in content:
            fence_string += fence_string[0]
        self.lit(fence_string)
        if info_string:
            self.out(info_string)
        self.cr()

    def html_inline(self, node, entering):
        self.lit(node.literal)

    def html_block(self, node, entering):
        self.lit('\n')
        self.emit_intented_literal(node.literal)
        self.emit_end_block(node)


# Define a new helper method that would be an in-place replacement
# for commonmark.commonmark.
def commonmark_to_html(markup):
    parser = commonmark.Parser()
    ast = parser.parse(markup)
    return PlainTextRenderer().render(ast)


if __name__ == "__main__":
    # Run the parser on STDIN and write to STDOUT.
    import sys
    print(commonmark_to_html(sys.stdin.read()))



import filecmp
import os
import shutil
import sys
from contextlib import ExitStack
from io import StringIO
from typing import List
from unittest import mock

import pytest
from recipemd.cli.main import Exit, main


@pytest.mark.parametrize(
    'type,dir,arguments',
    [
        ('valid', 'basic', []),
        ('valid', 'title', ['-t']),
        ('valid', 'ingredients', ['-i']),
        ('valid', 'json', ['-j']),
        ('valid', 'multiply', ['-m2']),
        ('valid', 'yield', ['-y', '10 servings']),
        ('valid', 'flatten', ['-f']),
        ('invalid', 'yield_no_matching', ['-y', '500 ml']),
        ('invalid', 'yield_required_invalid', ['-y', 'cheese']),
        ('invalid', 'multiply_unit', ['-m', '500 ml']),
        ('invalid', 'multiply_invalid', ['-m', 'cheese']),
    ]
)
def test_valid_args(type: str, dir: str, arguments: List[str]):
    test_dir = os.path.join(os.path.dirname(__file__), 'test_main', type, dir)
    input_file = os.path.join(test_dir, 'input.md')
    with ExitStack() as stack:
        stack.enter_context(mock.patch('sys.argv', [''] + arguments + [input_file]))
        stack.enter_context(mock.patch('sys.stdout', new_callable=StringIO))
        stack.enter_context(mock.patch('sys.stderr', new_callable=StringIO))
        if type == 'invalid':
            with pytest.raises(Exit):
                main()
        else:
            main()
        actual_stdout = sys.stdout.getvalue()  # type: ignore
        actual_stderr = sys.stderr.getvalue()  # type: ignore

    expected_stdout_file = os.path.join(test_dir, 'stdout')
    assert_equal_to_file_content(actual_stdout, expected_stdout_file)  # type: ignore

    expected_stderr_file = os.path.join(test_dir, 'stderr')
    assert_equal_to_file_content(actual_stderr, expected_stderr_file)  # type: ignore


def test_export_links():
    test_dir = os.path.join(os.path.dirname(__file__), 'test_main', 'export_links')
    input_file = os.path.join(test_dir, 'input.md')

    actual_output_dir = os.path.join(test_dir, 'actual')
    shutil.rmtree(actual_output_dir, ignore_errors=True)

    with ExitStack() as stack:
        actual_output_dir_relative = os.path.relpath(actual_output_dir, os.getcwd())
        stack.enter_context(mock.patch('sys.argv', ['', '--export-links', actual_output_dir_relative, input_file]))
        stack.enter_context(mock.patch('sys.stdout', new_callable=StringIO))
        stack.enter_context(mock.patch('sys.stderr', new_callable=StringIO))

        main()

        actual_stdout = sys.stdout.getvalue()  # type: ignore
        actual_stderr = sys.stderr.getvalue()  # type: ignore

    expected_stdout_file = os.path.join(test_dir, 'stdout')
    assert_equal_to_file_content(actual_stdout, expected_stdout_file)  # type: ignore

    expected_stderr_file = os.path.join(test_dir, 'stderr')
    if os.path.sep != '/':
        actual_stderr = actual_stderr.replace(os.path.sep, '/')  # type: ignore
    assert_equal_to_file_content(actual_stderr, expected_stderr_file)  # type: ignore

    expected_output_dir = os.path.join(test_dir, 'expected')
    files_to_compare = set(os.listdir(expected_output_dir) + os.listdir(actual_output_dir))
    (_, mismatch, errors) = filecmp.cmpfiles(expected_output_dir, actual_output_dir, files_to_compare, shallow=False)
    assert len(mismatch) == 0
    assert len(errors) == 0

    shutil.rmtree(actual_output_dir, ignore_errors=True)


def assert_equal_to_file_content(actual_str, expected_file):
    if os.environ.get('UPDATE_SNAPSHOTS'):
        if actual_str == '':
            try:
                os.remove(expected_file)
            except:
                pass
        else:
            with open(expected_file, 'w', encoding='UTF-8') as f:
                f.write(actual_str)

    try:
        with open(expected_file, 'r', encoding='UTF-8') as f:
            expected_str = f.read()
    except:
        expected_str = ''
    assert actual_str == expected_str



import setuptools

from recipemd import __version__

tests_require = [
    'pytest==6.2.5',
    'pytest-cov==2.8.1',
    'tox==3.20.1',
]

docs_require = [
    'Sphinx==2.2.2',
    'm2r==0.2.1',
    # restrict version of m2r depencency, see https://github.com/Azure/azure-sdk-for-python/issues/22019#issuecomment-987195928
    'mistune<2.0.0', 
    'sphinxcontrib.fulltoc==1.2.0',
    'sphinxcontrib.autoprogram==0.1.5',
    'sphinx_autodoc_typehints==1.10.3',
    'sphinxcontrib.apidoc==0.3.0',
    'sphinx-autobuild==0.7.1'
]

release_requires = [
    'twine==3.1.1',
    'wheel'
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="recipemd",
    version=__version__,
    author="Tilman Stehr",
    author_email="tilman@tilman.ninja",
    description="Markdown recipe manager, reference implementation of RecipeMD",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://recipemd.org",
    packages=setuptools.find_packages(),
    python_requires='>=3.7,<4',
    install_requires=[
        'dataclasses-json>=0.5.2,<0.6.0',
        'yarl>=1.3.0,<2.0.0',
        # commonmarkextension has to be vendorized due to #28. This can be removed once GovReady/CommonMark-py-Extensions#5  or we
        # remove or dependency on commonmarkextensions altogether
        # 'commonmarkextensions==0.0.5',
        'commonmark>=0.9.1,<1.0.0',
        'argcomplete>=1.10.0,<2.0.0',
        'pyparsing~=2.4.2',
    ],
    extras_require={
        'tests': tests_require,
        'docs': docs_require,
        'release': release_requires,
        'dev': tests_require+docs_require+release_requires
    },
    entry_points={
        'console_scripts': [
            'recipemd=recipemd.cli.main:main',
            'recipemd-find=recipemd.cli.find:main',
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
        "Topic :: Text Processing :: Markup",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
    ],
    project_urls={
        'Documentation': 'https://recipemd.org/reference_implementation.html',
        'Source': 'https://github.com/tstehr/recipemd',
    },
)



import codecs
import os.path
import re
import sys

import commonmark
import commonmark.render.html
import recipemd._vendor.commonmark_extensions.plaintext

def get_tests():
    if not os.path.exists("spec.txt"):
        print("Download the latest CommonMark spec.txt, e.g.:")
        print("wget https://raw.githubusercontent.com/jgm/CommonMark/master/spec.txt")
        sys.exit(0)

    # Read the spec.txt file and normalize line endings (I'm copying from CommonMark-py here).
    with codecs.open("spec.txt", encoding="utf-8") as f:
        spec = "".join(f.readlines())
    spec = spec.replace("→", "\t")

    # Chop off part that indicates end of tests.
    spec = re.sub(re.compile("^<!-- END TESTS -->(.|[\n])*", flags=re.M), '', spec)

    # Find all examples embedded in the spec, which serve as unit tests. I'm copying from:
    # https://github.com/rtfd/CommonMark-py/blob/master/CommonMark/tests/run_spec_tests.py
    tests = re.findall(
        re.compile(
            "^`{32} example\n"
            "([\s\S]*?)^\.\n([\s\S]*?)"
            "^`{32}$"
            "|^#{1,6} *(.*)$",
            re.M),
        spec)

    # Re-form into Python dicts.
    current_section = [None] # wrap in object to make closure
    def is_section_name(match):
        if not match[2] == "":
            current_section[0] = match[2]
            return True
        return False
    return [
        {
            'markdown': match[0],
            'html': match[1],
            'section': current_section[0],
            'number': example_number,
        }
        for example_number, match
        in enumerate(tests)
        if not is_section_name(match)
    ]

def run_tests():
    test_output = []

    parser = commonmark.Parser()

    for test in get_tests():
        # Render as text and output.
        # Since we don't have reference output, we'll make our own and
        # just track if it changes.

        renderer = recipemd._vendor.commonmark_extensions.plaintext.PlainTextRenderer()
        ast = parser.parse(test['markdown'])

        heading = "TEST (%s)" % test['section']
        print(heading)
        print("=" * len(heading))
        print()
        print("markdown:")
        print(test['markdown'])
        print("HTML:")
        print(test['html'])
        print("output:")

        try:
            output = renderer.render(ast)
        except recipemd._vendor.commonmark_extensions.plaintext.RawHtmlNotAllowed:
            print("[raw HTML is not permitted in plain text output]\n")
        else:
            print(output)

        # Render as CommonMark, then re-parse it, render that to
        # HTML, and see if the HTML matches the reference HTML.
        cm = recipemd._vendor.commonmark_extensions.plaintext.CommonMarkToCommonMarkRenderer().render(ast)
        ast = parser.parse(cm)
        html = commonmark.render.html.HtmlRenderer().render(ast)
        if html != test['html']:
            # This is an error. Round-tripping didn't work.
            print("Round-tripping CommonMark failed. It generated:")
            print(cm)


run_tests()




import unittest

import commonmark
from recipemd._vendor.commonmark_extensions.tables import ParserWithTables, RendererWithTables
from recipemd._vendor.commonmark_extensions.plaintext import PlainTextRenderer


class GithubFlavoredTablesTests(unittest.TestCase):
    """Test against Github-flavored Markdown defined at
    https://github.github.com/gfm/#tables-extension-
    Unfortunately many of the tests --- especially tests
    for invalid table markup --- are failing."""


    def assertRender(self, markdown, expected):
        parser = ParserWithTables(options={"multiline_table_cells": False})
        ast = parser.parse(markdown)
        html = RendererWithTables().render(ast)
        html = html.rstrip()
        self.assertEqual(html, expected)


    def test_simple(self):
        self.assertRender(
"""| foo | bar |
| --- | --- |
| baz | bim |""",

"""<table>
<thead>
<tr>
<th>foo</th>
<th>bar</th>
</tr>
</thead>
<tbody>
<tr>
<td>baz</td>
<td>bim</td>
</tr>
</tbody>
</table>"""            
        )


    def test_alignment(self):
        self.assertRender(
"""| abc | defghi |
:-: | -----------:
bar | baz""",

"""<table>
<thead>
<tr>
<th align="center">abc</th>
<th align="right">defghi</th>
</tr>
</thead>
<tbody>
<tr>
<td align="center">bar</td>
<td align="right">baz</td>
</tr>
</tbody>
</table>""")


    def test_escapes(self):
        self.assertRender(
"""| f\|oo  |
| ------ |
| b `\|` az |
| b **\|** im |""",

"""<table>
<thead>
<tr>
<th>f|oo</th>
</tr>
</thead>
<tbody>
<tr>
<td>b <code>|</code> az</td>
</tr>
<tr>
<td>b <strong>|</strong> im</td>
</tr>
</tbody>
</table>""")


    def test_break(self):
        self.assertRender(
"""| abc | def |
| --- | --- |
| bar | baz |
> bar""",

"""<table>
<thead>
<tr>
<th>abc</th>
<th>def</th>
</tr>
</thead>
<tbody>
<tr>
<td>bar</td>
<td>baz</td>
</tr>
</tbody>
</table>

<blockquote>
<p>bar</p>
</blockquote>""")


    def test_break2(self):
        self.assertRender(
"""| abc | def |
| --- | --- |
| bar | baz |
bar""",

"""<table>
<thead>
<tr>
<th>abc</th>
<th>def</th>
</tr>
</thead>
<tbody>
<tr>
<td>bar</td>
<td>baz</td>
</tr>
<tr>
<td>bar</td>
<td></td>
</tr>
</tbody>
</table>
<p>bar</p>""")


    def test_header_cells_count_matches_delimiter_row(self):
        self.assertRender(
"""| abc | def |
| --- |
| bar |""",

"""<p>| abc | def |
| --- |
| bar |</p>""")


    def test_cols_need_not_match(self):
        self.assertRender(
"""| abc | def |
| --- | --- |
| bar |
| bar | baz | boo |""",

"""<table>
<thead>
<tr>
<th>abc</th>
<th>def</th>
</tr>
</thead>
<tbody>
<tr>
<td>bar</td>
<td></td>
</tr>
<tr>
<td>bar</td>
<td>baz</td>
</tr>
</tbody>
</table>""")


    def test_no_tbody_if_empty(self):
        self.assertRender(
"""| abc | def |
| --- | --- |""",

"""<table>
<thead>
<tr>
<th>abc</th>
<th>def</th>
</tr>
</thead>
</table>""")

            
class MultilineTablesTests(unittest.TestCase):
    """Test the multiline cell mode."""


    def assertRender(self, markdown, expected):
        parser = ParserWithTables()
        ast = parser.parse(markdown)
        html = RendererWithTables().render(ast)
        html = html.rstrip()
        self.assertEqual(html, expected)


    def test_simple(self):
        self.assertRender(
"""| foo | bar |
| === | === |
| baz | bim |
| baz | bim |""",

"""<table>
<thead>
<tr>
<th>foo</th>
<th>bar</th>
</tr>
</thead>
<tbody>
<tr>
<td><p>baz
baz</p>
</td>
<td><p>bim
bim</p>
</td>
</tr>
</tbody>
</table>"""            
        )
        

class PlainTextRendererTests(unittest.TestCase):
    """Test the PlainTextRenderer cell mode."""


    def assertRender(self, markdown, expected):
        parser = commonmark.Parser()
        ast = parser.parse(markdown)
        text = PlainTextRenderer().render(ast).rstrip()
        self.assertEqual(text, expected)


    def test_simple(self):
        self.assertRender("""Hello""", """Hello""")

    def test_heading(self):
        self.assertRender("""# Heading *Formatting*""", """Heading *Formatting*\n####################""")

if __name__ == '__main__':
    unittest.main()